{------------------------------------------------------------------------------}
{ TWAIN source code:                                                           }
{ Copyright (C) '91-'92 TWAIN Working Group:                                   }
{ Aldus, Caere, Eastman-Kodak, Logitech,                                       }
{ Hewlett-Packard Corporations.                                                }
{ All rights reserved.                                                         }
{                                                                              }
{ This source code was converted from C to Turbo Pascal for Windoes by         }
{                                                                              }
{   MCM Design                                                                 }
{   Marc Christopher Martin                                                    }
{   Australiensvej 34                                                          }
{   2100 Copenhagen OE                                                         }
{   Denmark                                                                    }
{                                                                              }
{   e-mail support@mcm-design.dk                                               }
{                                                                              }
{------------------------------------------------------------------------------}

Unit twa_glue;

{------------------------------------------------------------------------------}
{$A+} { Align data Byte = $A-, Word = $A+.                                     }
{$F+} { Force FAR call.                                                        }
{$G+} { Generate 80286 code.                                                   }
{$N-} { CoProcessor code generation.                                           }
{$R-} { Range check.                                                           }
{$W-} { Protected mode only $W-, Real mode stack frame support $W+.            }
{$X+} { Extended syntax (When on, function can be used as procedures).         }
{------------------------------------------------------------------------------}

interface

uses WinTypes, WinProcs,
     twain, twa_apph;

{ $DEFINE DEBUG}         { Show errors that result from application             }
                        { calls to Source Manager                              }

{$DEFINE GRAY_MENUS}    { gray's menu items to prevent a second call           }
                        { to Acquire or Select while the first call            }
                        { is still pending.                                    }
                        { Including the menu gray code will introduce          }
                        { a requirement in your application to support         }
                        { a similar menu scheme.                               }


{ Globals to this module ONLY                                                  }
var
  TWDSMOpen   : bool;             { Init to False.                             }
                                  { Glue code flag for an Open Source Manager  }
  hDSMDLL     : THandle;          { Init to 0.                 }
                                  { Handle to Source Manager for explicit load }
  hGlobalWnd  : hWnd;             { Global for window handle                   }
  TWDSOpen    : bool;             { Init to False.                             }
  TWDSEnabled : bool;             { Glue code flag for an Open Source          }
  TWUI        : TW_USERINTERFACE; { Fix as per Courisimault: 940518 - bd       }

{ Globals                                                                      }
  appID       : TW_IDENTITY;      { storage for App and DS (Source) states     }
  dsID        : TW_IDENTITY;
{  NewDSIdentity : TW_IDENTITY; }
  lpDSM_Entry : TDSM_Entry_Proc;  { entry point to the SM                      }


const  Valid_Handle = 32;         { valid windows handle SB >= 32              }

const TWItemSize  : array[0..12] of integer = (SizeOf(TW_INT8),
                                               SizeOf(TW_INT16),
	                                       SizeOf(TW_INT32),
	                                       SizeOf(TW_UINT8),
	                                       SizeOf(TW_UINT16),
	                                       SizeOf(TW_UINT32),
	                                       SizeOf(TW_BOOL),
	                                       SizeOf(TW_FIX32),
	                                       SizeOf(TW_FRAME),
	                                       SizeOf(TW_STR32),
	                                       SizeOf(TW_STR64),
	                                       SizeOf(TW_STR128),
	                                       SizeOf(TW_STR255));

{------------------------------------------------------------------------------}
{ Function prototypes from module DCA_GLUE.PAS                                 }
{------------------------------------------------------------------------------}

{ Candy routines                                                               }

procedure TWInitialize(pIdentity : pTW_IDENTITY; hMainWnd : hWnd);

procedure TWTerminate(hBmp : THandle);

{ Routines for DSM                                                             }

function  TWOpenDSM   : bool;

function  TWCloseDSM(hBmp : THandle) : bool;

function  TWIsDSMOpen : bool;

{ Routines for DS                                                              }

function  TWOpenDS    : bool;

function  TWCloseDS : bool;

function  TWEnableDS(ShowUI : TW_BOOL) : bool;

function  TWDisableDS : bool;

function  TWIsDSOpen  : bool;

function  TWIsDSEnabled : bool;

function  TWSelectDS  : bool;

{ MCM, Functions not originally included in the dca_glue file.                 }
{      They were included in version 1.0.                                      }

function  TWGetFileType : TW_UINT32;

function  TWGetPixelType : TW_UINT32;

function  TWSetPixelType(ItemType : TW_UINT16; ItemValue : TW_INT32) : TW_UINT16;

{ MCM, Routines for Errors/Interface                                           }
procedure Explain_Error(hTmpWnd : hWnd; State : PChar);

procedure GreyMenu(Flag : boolean);


implementation

uses twa_acq, twa_capt, twd_com,
     SysUtils, Menus, Classes, Forms, Controls, 
     StdCtrls, Dialogs, Buttons, Messages, ExtCtrls;

{------------------------------------------------------------------------------}
{ Defines for the current_state table used in Explain_Error                    }
{------------------------------------------------------------------------------}

const
  E_CLOSEDSM          : PChar = 'Error Closing DSM.';
  E_CLOSEDS           : PChar = 'Error Closing DS.';
  E_USERSELECT        : PChar = 'Error Accessing DS.';
  E_SETUPMEMXFER      : PChar = 'Error Setting up memory transfer.';
  E_DISABLEDS         : PChar = 'Error Disabling DS.';
  E_ENABLEDS          : PChar = 'Error Enabling DS.';
  E_GETFIRST          : PChar = 'Error Getting first Data Source.';
  E_GETNEXT           : PChar = 'Error Getting Next Data Source.';
  E_CAPPIXELGET       : PChar = 'Error Getting Cap Pixel Type.';
  E_CAPPIXELSET       : PChar = 'Error Setting Cap Pixel Type.';
  E_CAPFILEGET        : PChar = 'Error Getting Cap File Type.';
  E_CAPABILITY        : PChar = 'Error Setting Capability.';


{------------------------------------------------------------------------------}
{ FUNCTION: TWInitialize                                                       }
{                                                                              }
{ ARGS:    pIdentity information about the App                                 }
{          hMainWnd handle to App main window                                  }
{                                                                              }
{ RETURNS: none                                                                }
{                                                                              }
{ NOTES:   This simple copy to a static structure, appID, does not support     }
{          multiple connections. TODO, upgrade.                                }
{                                                                              }
{------------------------------------------------------------------------------}

procedure TWInitialize(pIdentity : pTW_IDENTITY;
                       hMainWnd  : hWnd);
begin
  TWDSMOpen    := False;      { Added by MCM.                                  }
  hDSMDLL      := 0;          { Added by MCM.                                  }
  TWDSOpen     := False;      { Added by MCM.                                  }
  lpDSM_Entry  := Nil;        { Added by MCM.                                  }

  appID        := pIdentity^; { Get copy of app info                           }
  hGlobalWnd   := hMainWnd;   { Get copy of app window handle                  }
end; {End TWInitialize.                                                        }


{------------------------------------------------------------------------------}
{ FUNCTION:    TWTerminate                                                     }
{                                                                              }
{ ARGS:    none                                                                }
{                                                                              }
{ RETURNS: none                                                                }
{                                                                              }
{ NOTES:    To properly back out of a connection to a Source:                  }
{                TWDisableDS      disable the Source                           }
{                TWCloseDS        close the Source                             }
{                TWCloseDSM       close the Source Manager                     }
{                                                                              }
{ The simple parameterless calls do not support multiple connections.          }
{                                                                              }
{ MCM, This procedure is not part of the original TWAIN source code.           }
{                                                                              }
{------------------------------------------------------------------------------}

procedure TWTerminate(hBmp : THandle);
begin
  { Always try to take SM and Source down with you.                            }
  if TWIsDSEnabled
  then if TWDisableDS
       then {CheckSpecialMenu(hWnd, TW_APP_CLOSEDS)};
  if TWIsDSOpen
  then if TWCloseDS
       then {CheckSpecialMenu(hWnd, TW_APP_DISABLE)};
  if TWIsDSMOpen
  then if TWCloseDSM(hBmp)
       then {CheckSpecialMenu(hWnd, TW_APP_CLOSESM)};
end; { End TWTerminate.                                                        }


{------------------------------------------------------------------------------}
{ FUNCTION: TWOpenDSM                                                          }
{                                                                              }
{ ARGS:    none                                                                }
{                                                                              }
{ RETURNS: current state of the Source Manager                                 }
{                                                                              }
{ NOTES:     1). be sure SM is not already open                                }
{            2). explicitly load the .DLL for the Source Manager               }
{            3). call Source Manager to:                                       }
{                - opens/loads the SM                                          }
{                - pass the handle to the app's window to the SM               }
{                - get the SM assigned appID.id field                          }
{                                                                              }
{------------------------------------------------------------------------------}

function TWOpenDSM : bool;
const WINDIRPATHSIZE = 160;
var   twRC           : TW_UINT16;
   {$IFNDEF WIN32}
      OpenFiles      : TOFStruct;
   {$ENDIF}
      WinDir         : array[0..WINDIRPATHSIZE] of char;
      DSMName        : TW_STR32;
begin
  { Debug only.                                                                }
  {$IFDEF DEBUG}
    if TWDSMOpen
    then LogMessage('DSM already open');
  {$ENDIF}

  { Only open SM if currently closed                                           }
  if Not(TWDSMOpen)
  then begin
       { Open the SM, Refer explicitly to the library so we can post a nice    }
       { message to the the user in place of the "Insert TWAIN.DLL in drive    }
       { A:" posted by Windows if the SM is not found.                         }

       GetWindowsDirectory(WinDir, WINDIRPATHSIZE);
       { Hardcode to seperate dca_main.c & dca_glue.c more completely          }
       { lstrcpy (DSMName, "TWAIN.DLL");                                       }

       { NOTE: could use hInst from elsewhere and call name in from resource   }
{       LoadString(GetClassLong(hGlobalWnd, GCL_HMODULE),
                   IDS_DSMNAME,
                   DSMName,
                   sizeof(TW_STR32)); }

       {$IFDEF WIN32}
         lstrcpy(DSMName, 'TWAIN_32.DLL');
       {$ELSE}
         lstrcpy(DSMName, 'TWAIN.DLL');
       {$ENDIF}
       if (WinDir[StrLen(WinDir)-1] <> '\')
       then lstrcat(WinDir, '\');
       lstrcat(WinDir, DSMName);
       
       {$IFDEF WIN32}
         if FileExists(WinDir)
       {$ELSE}
         if (OpenFile(WinDir, OpenFiles, OF_EXIST) <> -1)
       {$ENDIF}
       then begin
            hDSMDLL := LoadLibrary(WinDir);
            if (hDSMDLL >= VALID_HANDLE) { >= VALID_HANDLE = 32                }
            then begin
                 lpDSM_Entry := TDSM_Entry_Proc(GetProcAddress(hDSMDLL, MAKEINTRESOURCE(1))); { 'DSM_ENTRY' }
                 if (@lpDSM_Entry <> Nil)
                 then begin
                 { This call performs four important functions:                }
                 { - opens/loads the SM                                        }
                 { - passes the handle to the app's window to the SM           }
                 { - returns the SM assigned appID.id field                    }
                 { - be sure to test the return code for SUCCESSful open of SM }
                      twRC := lpDSM_Entry(@appID,
                                          Nil,
                                          DG_CONTROL,
                                          DAT_PARENT,
                                          MSG_OPENDSM,
                                          TW_MEMREF(@hGlobalWnd));
                      case twRC of
                      { Needed for house keeping.  Do single open and do not   }
                      { close SM which is not already open ....                }
                      TWRC_SUCCESS : begin
                                       TWDSMOpen := True;
                                       if (MessageLevel >= ML_FULL)
                                       then ShowRC_CC(hGlobalWnd, 0, 0, 0,
		    				      'Source Manager was Opened successfully',
		    				      'TWAIN Information');
                                     end;
                      { Trouble opening the SM, inform the user                }
                      TWRC_FAILURE : begin
                                       TWDSMOpen := False;
                                       LogMessage('OpenDSM failure');
	        		       if (MessageLevel >= ML_ERROR)
	    	                       then ShowRC_CC(hGlobalWnd, 1, twRC, 0,
        	        		              '',
            	    		                      'DG_CONTROL / DAT_PARENT / MSG_OPENDSM');
                                     end;
                      else           begin
                                       TWDSMOpen := False;
                                       LogMessage('OpenDSM failure');
	        		       if (MessageLevel >= ML_ERROR)
	    	                       then ShowRC_CC(hGlobalWnd, 1, twRC, 0,
        	        		              '',
            	    		                      'DG_CONTROL / DAT_PARENT / MSG_OPENDSM');
                                     end;
                      end;
                 end
                 else begin { Could not obtain address to DSM_ENTRY.              }
                      if (MessageLevel >= ML_ERROR)
                      then ShowRC_CC(hGlobalWnd, 0, 0, 0,
                                     'Could not obtain address to DSM_ENTRY.',
                                     {$IFDEF WIN32}
                                       'TWAIN_32.DLL');
                                     {$ELSE}
                	               'TWAIN.DLL');
                                     {$ENDIF}
                 end;
            end
            else begin { Could not load library.                               }
                 if (MessageLevel >= ML_ERROR)
                 then ShowRC_CC(hGlobalWnd, 0, 0, 0,
                                'Could not load library.',
                                {$IFDEF WIN32}
                                  'TWAIN_32.DLL');
                                {$ELSE}
                	          'TWAIN.DLL');
                                {$ENDIF}
            end;
       end
       else begin
            if (MessageLevel >= ML_ERROR)
            then ShowRC_CC(hGlobalWnd, 0, 0, 0,
                           'Twain DLL may not exist.',
                           {$IFDEF WIN32}
                             'TWAIN_32.DLL');
                           {$ELSE}
                	     'TWAIN.DLL');
                           {$ENDIF}
       end;
  end;

  { Let the caller know what happened.                                         }
  TWOpenDSM := TWDSMOpen;
end; {End TWOpenDSM.                                                           }


{------------------------------------------------------------------------------}
{ FUNCTION: TWCloseDSM                                                         }
{                                                                              }
{ ARGS:    none                                                                }
{                                                                              }
{ RETURNS: current state of Source Manager                                     }
{                                                                              }
{ NOTES:    1). be sure SM is already open                                     }
{           2). call Source Manager to:                                        }
{               - request closure of the Source identified by appID info       }
{                                                                              }
{------------------------------------------------------------------------------}

function TWCloseDSM(hBmp : THandle) : bool;
var twRC : TW_UINT16;
    Buffer : array[0..64] of char;
begin
  twRC := TWRC_FAILURE;
  if TWDSOpen
  then begin { A Source is Currently Open", "Cannot Close Source Manager. }
       if (MessageLevel  >= ML_ERROR)
       then ShowRC_CC(hGlobalWnd, 0, 0, 0,
                      'A Source is Currently Open',
                      'Cannot Close Source Manager');

  end
  else begin
       { Only close something which is already open                                 }
       if TWDSMOpen
       then begin
       { This call performs one important function:                            }
       { - tells the SM which application, appID.id, is requesting SM to close.}
       { - be sure to test return code, failure indicates SM did not close !!  }
            twRC := lpDSM_Entry(@appID,
                                Nil,
                                DG_CONTROL,
                                DAT_PARENT,
                                MSG_CLOSEDSM,
                                TW_MEMREF(@hGlobalWnd));

            if (twRC <> TWRC_SUCCESS)
            then begin
                 { Trouble closing the SM, inform the user.                    }
		 if (MessageLevel >= ML_ERROR)
		 then ShowRC_CC(hGlobalWnd, 1, twRC, 0,
		                '',
		                'DG_CONTROL / DAT_PARENT / MSG_CLOSEDSM');

                 StrFmt(Buffer, 'CloseDSM failure - twRC = %d', [twRC]);
		 LogMessage(buffer);
            end
            else begin
                 TWDSMOpen := False;
                 { Explicitly free the SM library.                             }
                 if (hDSMDLL > 0)
                 then begin
                     FreeLibrary(hDSMDLL);
                     hDSMDLL := 0;
                     { the data source id will no longer be valid after        }
                     { twain is killed.  If the id is left around the          }
                     { data source can not be found or opened                  }
                     dsID.Id := 0;
                 end;
		 if (MessageLevel >= ML_FULL)
		 then ShowRC_CC(hGlobalWnd, 0, 0, 0,
		                'Source Manager was Closed successfully',
			    	'TWAIN Information');

            end;
       end
       else begin { Cannot close Source Manager, Source Manager is Closed.     }
	    if (MessageLevel >= ML_ERROR)
	    then ShowRC_CC(hGlobalWnd, 0, 0, 0,
			   'Cannot Close Source Manager Source Manager Not Open',
			   'Sequence Error');
       end;
  end;
  { Let the caller know what happened.                                         }
  TWCloseDSM := (twRC = TWRC_SUCCESS);
end; { End TWCloseDSM.                                                         }


{------------------------------------------------------------------------------}
{ FUNCTION: TWIsDSMOpen                                                        }
{                                                                              }
{ ARGS:    none                                                                }
{                                                                              }
{ RETURNS: current state of Source Manager (open/closed)                       }
{                                                                              }
{ NOTES:   Just a way to reduce the number of global vars and keep the         }
{          state of SM information local to this module.  Let the caller,      }
{          app, know current state of the SM.                                  }
{------------------------------------------------------------------------------}

function TWIsDSMOpen : bool;
begin
  TWIsDSMOpen := TWDSMOpen;
end; { End TWIsDSMOpen.                                                        }


{------------------------------------------------------------------------------}
{ FUNCTION: TWOpenDS                                                           }
{                                                                              }
{ ARGS:    none                                                                }
{                                                                              }
{ RETURNS: current state of select Source                                      }
{                                                                              }
{ NOTES:    1). only attempt to open a source if it is currently closed        }
{           2). call Source Manager to:                                        }
{                - open the Source indicated by info in dsID                   }
{                - SM will fill in the unique .Id field                        }
{------------------------------------------------------------------------------}

function TWOpenDS : bool;
var twRC : TW_UINT16;
begin
  if Not(TWDSMOpen)
  then begin { DSM not open - cannot open DS.                                  }
       LogMessage('DSM not open - cannot open DS');
       if (MessageLevel >= ML_ERROR)
       then ShowRC_CC(hGlobalWnd, 0, 0, 0,
 		      'Cannot Open Source. Source Manager not Open',
 		      'TWAIN Error');
  end
  else begin { Source Manager is open.                                         }
       if Not(TWDSOpen)
       then begin
            { This will open the Source.                                       }
            twRC := lpDSM_Entry(@appID,
                                Nil,
                                DG_CONTROL,
                                DAT_IDENTITY,
                                MSG_OPENDS,
                                TW_MEMREF(@dsID));

            case twRC of
            TWRC_SUCCESS : begin
                             { do not change flag unless we successfully open. }
                             TWDSOpen := True;
			     LogMessage('OpenDS success');
		             if (MessageLevel >= ML_INFO)
		             then ShowTW_ID(hGlobalWnd, dsID,
	                		  'DG_CONTROL / DAT_IDENTITY / MSG_OPENDS TWRC_SUCCESS');
                           end;
            TWRC_FAILURE : begin
			     LogMessage('OpenDS failure');

			     { Trouble opening the Source                      }
			     { Determine Condition Code                        }
			     if (MessageLevel >= ML_ERROR) { Source Manager RC }
		       	     then ShowRC_CC(hGlobalWnd, 1, twRC, 0,
                                            '',
			                    'DG_CONTROL / DAT_IDENTITY / MSG_OPENDS');

			     { this is a patch required in the event that      }
                             { TWAcquire fails due to bad or non-existant      }
                             { during run of a autotest. The tests are         }
                             { contnued via the WM_PAINT msg.                  }
			     { InvalidateRect(NULL,NULL,TRUE);                 }
                           end;
            else           begin { Trouble opening the Source, unknown error.  }
	                     if (MessageLevel >= ML_ERROR)
		       	     then ShowRC_CC(hGlobalWnd, 0, 0, 0,
					    'Unknown Return Code',
					    'DG_CONTROL / DAT_IDENTITY / MSG_OPENDS');
                           end;
            end;
       end
       else begin { Source is already open, TWAIN Information.                 }
            LogMessage('Source already open');
	    if (MessageLevel >= ML_FULL)
	    then ShowRC_CC(hGlobalWnd, 0, 0, 0,
 			   'Source is already open',
                           'TWAIN Information');
       end;
  end;
  TWOpenDS := TWDSOpen;
end; { End TWOpenDS.                                                           }



{------------------------------------------------------------------------------}
{ FUNCTION: TWCloseDS                                                          }
{                                                                              }
{ ARGS:    none                                                                }
{                                                                              }
{ RETURNS: none                                                                }
{                                                                              }
{ NOTES:    1). only attempt to close an open Source                           }
{           2). call Source Manager to:                                        }
{                - ask that Source identified by info in dsID close itself     }
{------------------------------------------------------------------------------}

function TWCloseDS : bool;
var twRC : TW_UINT16;
begin
  twRC := TWRC_FAILURE;
  if TWDSOpen
  then begin
       if TWDSEnabled
       then begin { Source is Currently Enabled, Cannot Close Source.          }
            if (MessageLevel >= ML_ERROR)
	    then ShowRC_CC(hGlobalWnd, 0, 0, 0,
			   'Source is Currently Enabled',
                           'Cannot Close Source');
       end
       else begin
            { Close an open Source.                                            }
            twRC := lpDSM_Entry(@appID, Nil,
                                DG_CONTROL,
                                DAT_IDENTITY,
                                MSG_CLOSEDS,
                                TW_MEMREF(@dsID));
	    LogMessage('CloseDS');

            { show error on close.                                             }
            if (twRC <> TWRC_SUCCESS)
            then begin
	         if (MessageLevel >= ML_ERROR)
		 then ShowRC_CC(hGlobalWnd, 1, twRC, 0,
		                '',
		                'DG_CONTROL / DAT_IDENTITY / MSG_CLOSEDS');
            end
            else begin
                 TWDSOpen := False;
                 GreyMenu(False);
                 dsID.Id := 0;
                 dsID.ProductName[0] := #0;
		 if (MessageLevel >= ML_FULL)
		 then ShowRC_CC(hGlobalWnd, 0, 0, 0,
				'Source was Closed successfully',
				'TWAIN Information');
            end;
       end;
  end
  else begin { Cannot Close Source Source Not Open, Sequence Error.           }
       if (MessageLevel >= ML_ERROR)
       then ShowRC_CC(hGlobalWnd, 0, 0, 0,
		      'Cannot Close Source. Source Not Open',
		      'Sequence Error');
  end;
  TWCloseDS := (twRC = TWRC_SUCCESS);
end; { End TWCloseDS.                                                          }


{------------------------------------------------------------------------------}
{ FUNCTION: TWEnableDS                                                         }
{                                                                              }
{ ARGS:    none                                                                }
{                                                                              }
{ RETURNS: BOOL for True=open; False=not open/fail                             }
{                                                                              }
{ NOTES:    1). only enable an open Source                                     }
{           2). call the Source Manager to:                                    }
{                - bring up the Source's User Interface                        }
{------------------------------------------------------------------------------}

function TWEnableDS(ShowUI : TW_BOOL) : bool;
var twRC   : TW_UINT16;
    twUI   : TW_USERINTERFACE;
  {$IFDEF VER70}
    Result : bool;
  {$ENDIF}
begin
  Result := False;

  { Only enable open Source's.                                                 }
  if TWDSOpen
  then begin
       if TWDSEnabled
       then begin { Cannot Enable Source, already enabled, TWAIN Error.        }
            if (MessageLevel >= ML_ERROR)
	    then ShowRC_CC(hGlobalWnd, 0, 0, 0,
			   'Cannot Enable Source, already enabled',
			   'TWAIN Error');
       end
       else begin
            { Grey out file menu items.                                        }
            GreyMenu(True);

            { This will display the Source User Interface. The Source should   }
            { only display a user interface that is compatible with the group  }
            { defined by appID.                                                }
            { SupportedGroups (in our case DG_IMAGE | DG_CONTROL)              }
            twUI.hParent := hGlobalWnd;
            twUI.ShowUI  := ShowUI;
            twRC := lpDSM_Entry(@appID,
                                @dsID,
                                DG_CONTROL,
                                DAT_USERINTERFACE,
                                MSG_ENABLEDS,
                                TW_MEMREF(@twUI));
            if (twRC <> TWRC_SUCCESS)
            then begin
	         if (MessageLevel >= ML_ERROR)
	         then ShowRC_CC(hGlobalWnd, 1, twRC, 1,
	                	'',
	                	'DG_CONTROL / DAT_USERINTERFACE / MSG_ENABLEDS');
            end
            else begin
                 Result := True;
                 TWDSEnabled := True;
	         if (MessageLevel >= ML_FULL)
	         then ShowRC_CC(hGlobalWnd, 0, 0, 0,
	                	'Source Enabled',
	                	'DG_CONTROL / DAT_USERINTERFACE / MSG_ENABLEDS');
            end;
       end;
  end
  else begin { Cannot Enable Source, No Source Open, TWAIN Error.              }
       if (MessageLevel >= ML_ERROR)
       then ShowRC_CC(hGlobalWnd, 0, 0, 0,
		      'Cannot Enable Source. No Source Open',
		      'TWAIN Error');
  end;
  TWEnableDS := Result;
end; { End TWEnableDS.                                                         }


{------------------------------------------------------------------------------}
{ FUNCTION: TWDisableDS                                                        }
{                                                                              }
{ ARGS:    none                                                                }
{                                                                              }
{ RETURNS: none                                                                }
{                                                                              }
{ NOTES:    1). only disable an open Source                                    }
{           2). call Source Manager to:                                        }
{                - ask Source to hide it's User Interface                      }
{------------------------------------------------------------------------------}

function TWDisableDS : bool;
var twRC   : TW_UINT16;
    twUI   : TW_USERINTERFACE;
begin
  twRC := TWRC_FAILURE;

  if Not(TWDSEnabled)
  then begin { Cannot Disable Source, Source Not Enabled, Sequence Error.      }
       if (MessageLevel >= ML_ERROR)
       then ShowRC_CC(hGlobalWnd, 0, 0, 0,
		      'Cannot Disable Source. Source Not Enabled',
		      'Sequence Error');
  end
  else begin
       { only disable open Source's.                                           }
       if (TWDSOpen = True)
       then begin
            { Hide the Source UI.                                              }
            twUI.hParent := hGlobalWnd;
            twUI.ShowUI  := TW_BOOL(TWON_DONTCARE8 <> 0);
            twRC := lpDSM_Entry(@appID, @dsID,
                                DG_CONTROL,
                                DAT_USERINTERFACE,
                                MSG_DISABLEDS,
                                TW_MEMREF(@twUI));
              if (twRC = TWRC_SUCCESS)
              then begin
        	   LogMessage('DS Disabled');
                   TWDSEnabled := False;
           	   if (MessageLevel >= ML_FULL)
                   then ShowRC_CC(hGlobalWnd, 0, 0, 0,
                	          'Source Disabled',
                	          'DG_CONTROL / DAT_USERINTERFACE / MSG_DISABLEDS');
              end
              else begin
        	   if (MessageLevel >= ML_ERROR)
                   then ShowRC_CC(hGlobalWnd, 1, twRC, 1,
               	                  '',
               	                  'DG_CONTROL / DAT_USERINTERFACE / MSG_DISABLEDS');
              end;
       end;
  end;
  TWDisableDS := (twRC = TWRC_SUCCESS);
end; { End TWDisableDS.                                                        }


{------------------------------------------------------------------------------}
{ FUNCTION: TWIsDSOpen                                                         }
{                                                                              }
{ ARGS:    none                                                                }
{                                                                              }
{ RETURNS: current state of the Source                                         }
{                                                                              }
{ NOTES:    Just a way to reduce the number of global vars and keep the        }
{           state of Source information local to this module.  Let the caller, }
{           app, know current state of the Source.                             }
{------------------------------------------------------------------------------}

function TWIsDSOpen : bool;
begin
  TWIsDSOpen := TWDSOpen;
end; { End TWIsDSOpen.                                                         }


{------------------------------------------------------------------------------}
{ FUNCTION: TWIsDSEnabled                                                      }
{                                                                              }
{ ARGS:    none                                                                }
{                                                                              }
{ RETURNS: current state of Source (Enabled?)                                  }
{                                                                              }
{ NOTES:   Just a way to reduce the number of global vars and keep the         }
{          state of SM information local to this module.  Let the caller,      }
{          app, know current state of the SM.                                  }
{                                                                              }
{------------------------------------------------------------------------------}
function TWIsDSEnabled : bool;
begin
  TWIsDSEnabled := TWDSEnabled;
end; { End TWIsDSEnabled.                                                      }


{------------------------------------------------------------------------------}
{ FUNCTION: DCSelectDS                                                         }
{                                                                              }
{ ARGS:    none                                                                }
{                                                                              }
{ RETURNS: twRC TWAIN status return code                                       }
{                                                                              }
{ NOTES:   1). call the Source Manager to:                                     }
{              - have the SM put up a list of the available Sources            }
{              - get information about the user selected Source from           }
{                NewDSIdentity, filled by Source                               }
{------------------------------------------------------------------------------}

function TWSelectDS : bool;
var twRC          : TW_UINT16;
    NewDSIdentity : TW_IDENTITY; 
begin
  if TWDSOpen
  then begin
       LogMessage('TWSelectDS - source already open');

       { A Source is already open, Close Source before Selecting a New Source. }
       if (MessageLevel >= ML_ERROR)
       then ShowRC_CC(hGlobalWnd, 0, 0, 0,
                      'A Source is already open.  Close Source before Selecting a New Source',
               	      'DG_CONTROL/DAT_IDENTITY/MSG_USERSELECT');
       twRC := TWRC_FAILURE;
  end
  else begin
       { I will settle for the system default.  Shouldn't I get a highlight    }
       { on system default without this call?                                  }
       twRC := lpDSM_Entry(@appID,
                           Nil,
                           DG_CONTROL,
                           DAT_IDENTITY,
                           MSG_GETDEFAULT,
                           TW_MEMREF(@NewDSIdentity));

       { This call performs one important function:                            }
       { - should cause SM to put up dialog box of available Source's          }
       { - tells the SM which application, appID.id, is requesting, REQUIRED   }
       { - returns the SM assigned NewDSIdentity.id field, you check if        }
       {   changed (needed to talk to a particular Data Source)                }
       { - be sure to test return code, failure indicates SM did not close !!  }
       if (twRC = TWRC_SUCCESS) or True
       then twRC := lpDSM_Entry(@appID,
                                Nil,
                                DG_CONTROL,
                                DAT_IDENTITY,
                                MSG_USERSELECT,
                                TW_MEMREF(@NewDSIdentity));


       { Check if the user changed the Source and react as apporpriate.        }
       { - TWRC_SUCCESS, log in new Source                                     }
       { - TWRC_CANCEL,  keep the current Source                               }
       { - default,      check the codes in a status message, display result   }
       case twRC of
       TWRC_SUCCESS : begin
                        dsID := NewDSIdentity;
                        if (MessageLevel >= ML_INFO)
		        then ShowTW_ID(hGlobalWnd, dsID,
	                	       'DG_CONTROL / DAT_IDENTITY / MSG_USERSELECT');
                      end;
       TWRC_CANCEL  : begin
	           	if (MessageLevel >= ML_INFO)
		        then ShowRC_CC(hGlobalWnd, 1, twRC, 0,
	                	       '',
	                	       'DG_CONTROL / DAT_IDENTITY / MSG_USERSELECT');
                      end;
       else           begin
	           	if (MessageLevel >= ML_ERROR)
		        then ShowRC_CC(hGlobalWnd, 1, twRC, 0,
	                	       '',
	                	       'DG_CONTROL / DAT_IDENTITY / MSG_USERSELECT');
                      end;
       end;
  end;
  { Let the caller know what happened.                                         }
  if (twRC <> 0)
  then TWSelectDS := True
  else TWSelectDS := False;
end; { End TWSelectDS.                                                         }


{------------------------------------------------------------------------------}
{ FUNCTION: GreyMenu                                                           }
{                                                                              }
{ ARGS:    flag    desired new state for flags                                 }
{                                                                              }
{ RETURNS: none                                                                }
{                                                                              }
{ NOTE:    one important "feature" of this routine is that by graying          }
{          the menu selections it locks out double hits on them by the user    }
{          while I am dealing with the first hit.                              }
{                                                                              }
{          Including the menu gray code will introduce a requirement in        }
{          your application to support a similar menu scheme.                  }
{------------------------------------------------------------------------------}

procedure GreyMenu(Flag : boolean);
var hMainMenu : hMenu;
begin
  { Includes for my the App specific menu labels.                              }
  hMainMenu := GetMenu(hGlobalWnd);
  if Flag
  then begin
       {$IFDEF GRAY_MENUS}
         EnableMenuItem(hMainMenu, TW_APP_ACQUIRE, MF_BYCOMMAND or MF_GRAYED);
         EnableMenuItem(hMainMenu, TW_APP_SELECT_SOURCE, MF_BYCOMMAND or MF_GRAYED);
       {$ENDIF}
  end
  else begin
       {$IFDEF GRAY_MENUS}
         EnableMenuItem(hMainMenu, TW_APP_ACQUIRE, MF_BYCOMMAND or MF_ENABLED);
         EnableMenuItem(hMainMenu, TW_APP_SELECT_SOURCE, MF_BYCOMMAND or MF_ENABLED);
       {$ENDIF}
  end;
end; { End GreyMenu.                                                           }


{------------------------------------------------------------------------------}
{ FUNCTION: Explain_Error                                                      }
{                                                                              }
{ ARGS:    hWnd    handle to a display window                                  }
{          *state  ptr to a character string from dca_glue.h which defines     }
{                  these error strings.  Iff state is NULL then no extra       }
{                  text, probably using the standard App error messages,       }
{                  (See dcastrg.rc, stringtable)                               }
{                                                                              }
{ RETURNS: none                                                                }
{                                                                              }
{ MCM, Old function                                                            }
{                                                                              }
{------------------------------------------------------------------------------}

procedure Explain_Error(hTmpWnd : HWND; State : PChar);
{$IFDEF DEBUG}
var twRC     : TW_UINT16;
    twStatus : TW_STATUS;
    Details1 : array[0..255] of char;
    Details2 : array[0..255] of char;
    code     : integer;
    Error    : boolean;
begin
  Error := True;
  code  := IDS_NODETAIL;
  { Put in detail of current operation.                                        }
  if (State <> Nil)
  then lstrcpy(Details1, state)
  else Details1[0] := #0;

  { Determine details of failure from SM.                                      }
  if (@lpDSM_Entry <> Nil)
  then begin
       twRC := lpDSM_Entry(@appID, Nil,
                           DG_CONTROL,
                           DAT_STATUS,
                           MSG_GET,
                           TW_MEMREF(@twStatus));

       if (twRC = TWRC_SUCCESS)
       then begin
            case twStatus.ConditionCode of
            TWCC_BADCAP         : code  := IDS_BADCAP;
            TWCC_BADPROTOCOL    : code  := IDS_BADPROTOCOL;
            TWCC_BADVALUE       : code  := IDS_BADVALUE;
            TWCC_BUMMER         : code  := IDS_BUMMER;
	    TWCC_OPERATIONERROR : Error := False;

            { Error in the DS or DSM which was already reported to the user.   }
            { The App should not be redundant.  The operation failed.          }
            TWCC_LOWMEMORY      : code  := IDS_LOWMEMORY;
	    TWCC_MAXCONNECTIONS : code  := IDS_MAXCONNECTIONS;
	    TWCC_NODS           : code  := IDS_NODS;
	    TWCC_SEQERROR       : code  := IDS_SEQERROR;
	    TWCC_SUCCESS        : code  := IDS_SUCCESS;
            else                  code  := IDS_DEFAULT;
            end;
       end
       else code := IDS_NODETAIL;
  end
  else code := IDS_NODSM;
  if Error
  then begin
       { Now send out the text.                                                }
       LoadString(GetClassLong(hTmpWnd, GCL_HMODULE), code, Details2, 255);

       { Show the error.                                                       }
       lstrcat(Details1, Details2);
       MessageBox(hTmpWnd, Details1, 'TWAIN Error', mb_OK);
  end;
{$ELSE}
begin
{$ENDIF}
end; { End Explain_Error.                                                      }


{------------------------------------------------------------------------------}
{ FUNCTION: TWGetFileType. MCM, from version 1.0                               }
{                                                                              }
{ ARGS:    none                                                                }
{                                                                              }
{ RETURNS: the supported file format.                                          }
{                                                                              }
{ NOTES: If more than one file format is supported, DCGetFileType will return  }
{        the first format in the container.                                    }
{------------------------------------------------------------------------------}

function TWGetFileType : TW_UINT32;
var twRC  : TW_UINT16;
    twCap : TW_CAPABILITY;
    back  : TW_UINT16;
begin
  back := $FFFF;

  { App ONLY fills in the CAP_ of interest.                                    }
  twCap.Cap        := ICAP_IMAGEFILEFORMAT;
  twCap.hContainer := 0;

  { Have Source build a container and fill it with the CAP_ of interest.       }
  twRC := lpDSM_Entry(@appID,
                      @dsID,
                      DG_CONTROL,
                      DAT_CAPABILITY,
                      MSG_GET,
                      TW_MEMREF(@twCap));

    if (twRC <> TWRC_SUCCESS)
    then begin
         if (MessageLevel >= ML_ERROR)
	 then ShowRC_CC(hGlobalWnd, 0, twRC, 0,
	                '',
	                'MSG_GET & ICAP_IMAGEFILEFORMAT');
    end;

  { Add code to do a switch on contype.  Then call the appropriate extract     }
  { routine from the common container code routines.                           }
  case twCap.ConType of
  TWON_ONEVALUE    : ExtractOneValue(@twCap, @back);
                     { get a single value from the enumeration container.      }
  TWON_ENUMERATION : ExtractEnumerationValue(@twCap, @back, 0);
  end;

  { App is ALWAYS responsible for cleaning up the container.                   }
  if (twCap.hContainer <> 0)
  then GlobalFree(twCap.hContainer);

  TWGetFileType := back;
end; { End TWGetFileType.                                                     }


{------------------------------------------------------------------------------}
{ FUNCTION: TWGetPixelType. MCM, from version 1.0                              }
{                                                                              }
{ ARGS:    none                                                                }
{                                                                              }
{ RETURNS: the current pixel type                                              }
{                                                                              }
{ NOTES:   1. selects the CAP type, ICAP_PIXELTYPE                             }
{          2. calls the Source Manager to:                                     }
{             - have Source build a container who's contents will be the       }
{               current pixel type of the selected Source                      }
{          3. use common lib routine to extract value from the container       }
{             type built, Source's choice, by the Source                       }
{------------------------------------------------------------------------------}

function TWGetPixelType : TW_UINT32;
var twRC  : TW_UINT16;
    twCap : TW_CAPABILITY;
    back  : TW_INT32;
begin
  back := 0;

  { App ONLY fills in the CAP_ of interest.                                    }
  twCap.Cap := ICAP_PIXELTYPE;

  { Have Source build a container and fill it with the CAP_ of interest.       }
  twRC := lpDSM_Entry(@appID,
                      @dsID,
                      DG_CONTROL,
                      DAT_CAPABILITY,
                      MSG_GET,
                      TW_MEMREF(@twCap));

    if (twRC <> TWRC_SUCCESS)
    then begin
         if (MessageLevel >= ML_ERROR)
	 then ShowRC_CC(hGlobalWnd, 0, twRC, 0,
	                '',
                        'MSG_GET & ICAP_PIXELTYPE');
    end;
{    Explain_Error(hGlobalWnd, E_CAPPIXELGET); }

  { Add code to do a switch on contype.  Then call the appropriate extract     }
  { routine from the common container code routines.                           }

  case twCap.ConType of
  TWON_ONEVALUE    : ExtractOneValue(@twCap, @back);
                     { get a single value from the enumeration container.      }
  TWON_ENUMERATION : ExtractEnumerationValue(@twCap, @back, 1);
  end;

  { App is ALWAYS responsible for cleaning up the container.                   }
  if (twCap.hContainer <> 0)
  then GlobalFree(twCap.hContainer);

  TWGetPixelType := back;
end; { End TWGetPixelType.                                                     }



{------------------------------------------------------------------------------}
{ FUNCTION: TWSetPixelType. MCM, from version 1.0                              }
{                                                                              }
{ ARGS:    ItemType    DCTY_xxx, type of the ItemValue                         }
{          ItemValue   the none DCPT_xxx constant for pixel type, see DC.H     }
{                                                                              }
{ RETURNS: dcRC TWAIN status return code                                       }
{                                                                              }
{ NOTES:   1). build up a container of type OneValue filled with ItemType      }
{              and ItemValue                                                   }
{          2). call Source Manager to:                                         }
{              - give the Source access to the pixel type you wish to set      }
{                the Source to, assumes you have asked the Source previously   }
{                if it can handle the pixel type you are now setting           }
{------------------------------------------------------------------------------}

function TWSetPixelType(ItemType : TW_UINT16; ItemValue : TW_INT32) : TW_UINT16;
var twCap : TW_CAPABILITY;
    twRC  : TW_UINT16;
begin
  twCap.Cap     := ICAP_PIXELTYPE; { id of cap you want.                       }
  twCap.ConType := TWON_ONEVALUE;  { container type.                           }

  { App must build the container & pass the container handle to the Source.    }
  BuildUpOneValue(@twCap, ItemType, @ItemValue, 0);

  { It is assumed that the Source will read the container NOW.                 }
  twRC := lpDSM_Entry(@appID,
                      @dsID,
                      DG_CONTROL,
                      DAT_CAPABILITY,
                      MSG_SET,
                      TW_MEMREF(@twCap));

    if (twRC <> TWRC_SUCCESS)
    then begin
         if (MessageLevel >= ML_ERROR)
	 then ShowRC_CC(hGlobalWnd, 0, twRC, 0,
	                '',
                        'MSG_SET & ICAP_PIXELTYPE');
    end;

  { NOTE: the App ALWAYS is required to Free the container.                    }
  if (twCap.hContainer <> 0)
  then GlobalFree(twCap.hContainer);

  TWSetPixelType := twRC;
end; { End TWSetPixelType.                                                     }


end. { End Unit twa_glue.                                                      }
